
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.31.0rc3',
    'branch_nick': 'build-1.31.0rc3',
    'revision_id': 'ce39a0bf87105ff50fcc2a9ca5da69d187523f39',
    'revno': 12460
}
